/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <assert.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_data_id.h"

#include "initng_last.h"

s_entry LAST = { "last", SET, 4 };


/* returns TRUE if all use deps are started */
static int check_last(active_h * service, e_a_status status)
{
    active_h *current = NULL;
    active_h **deps;
    active_h **deps_current;

    assert(service);
    assert(service->name);

    /* if we are trying to set status START_DEP_MET */
    if (status != START_DEP_MET)
        return (TRUE);

    /* And LAST is set. */
    if (!active_db_is(&LAST, service))
        return (TRUE);

    deps = active_db_dep_on_me(service);
    /*
     * walk the active db, check if there is any other
     * service that are still starting and not also
     * have the last flag set.
     */
    while_active_db(current)
    {
        /* not check ourself */
        if (current == service)
            continue;


        switch (current->a_status)
        {
            case LOADING:
            case STARTING:
            case WAITING_FOR_START_DEP:
            case START_DEP_MET:
                break;
            default:
                continue;                   /* while_active_db */
        }


        /* if this service also shud be started last, continue */
        if (active_db_is(&LAST, current))
        {
            continue;
        }

        /* 20051015 SaTaN0rX :
         * Problem:
         * daemon/agetty depends on daemon/agetty/tty, which one is last.
         * system/splash depends on system/splash/stop, which one is also last.
         *
         * the last plugin won't daemon/agetty/tty change it's status to START_DEP_MET, 
         * because system/splash is not started, does not depend on daemon/agetty/tty, and is not last.
         * it's the same with system/splash/stop and daemon/agetty.
         *
         * To compensate for this, we will now get the deps for current, and see if current has any
         * dependency on a LAST service
         */

        deps_current = active_db_dep_on(current);
        if (deps_current)
        {
            int i = 0;
            int continue_outer_loop = 0;

            while (deps_current[i])
            {
                D_("testing dep %s of service %s\n", deps_current[i]->name,
                   current->name);
                if (active_db_is(&LAST, deps_current[i]))
                {
                    continue_outer_loop = 1;
                    D_("service %s depends on %s, which is set last, continuing\n", current->name, deps_current[i]->name);
                    break;
                }
                i++;
            }
            free(deps_current);
            if (continue_outer_loop)
                continue;
        }

        /* check so this service, are not waiting for our service that
           shud be last, to be started */
        if (deps)
        {
            int i = 0;

            while (deps[i] && deps[i] != current)
                i++;
            if (deps[i] == current)
                continue;
        }
        D_("LAST: not allowing %s to start, cause %s is still starting and not LAST\n", service->name, current->name);
        /*
         * There is still services
         * starting, wont allow
         * a_status change!
         */
        free(deps);
        return (FALSE);

    }

    /* accept this change */
    free(deps);
    return (TRUE);
}


int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_sdt_add(&LAST);
    initng_add_hook(CONFIRM_ASTATUS_CHANGE, 80, &check_last);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_sdt_del(&LAST);
    initng_del_hook(CONFIRM_ASTATUS_CHANGE, &check_last);
}
